/*******************************************************************************
**	Chips & Technologies 655xx clock driver
********************************************************************************
**
   Copyright	1997		The GGI Project
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
********************************************************************************
**	This code was derived from the following sources of information:
**
**	[1] svgalib 1.2.11 sources
**	[2] official technical specifications (http:/*www.chips.com) */
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Stephan_Meyer
#define	DRIVER_NAME	"Chips 655xx clock driver"
#define	DRIVER_REV	"$Revision: 1.8 $"

#include <kgi/module.h>

#include "Chips/chips.h"

struct pll_limits
{
	kgi_uint		fref;	/* reference frequency		*/
	kgi_ratio	a;	/* generator dependent factor	*/
	kgi_range	fvco;	/* VCO frequency limits		*/
	kgi_range	p;	/* post-VCO divider range	*/
	kgi_range	mul;	/* multiplier value range	*/
	kgi_range	div;	/* divider value range		*/
};

struct pll_state
{
	kgi_uint	f;		/* produced frequency		*/
	kgi_uint	fvco;		/* VCO frequency		*/
	kgi_uint	p;		/* post-VCO divider		*/
	kgi_uint	mul;		/* multiplier value		*/
	kgi_uint	div;		/* divider value range		*/
	kgi_uint	psn;		/* 655xx specific 0->1 1->4	*/
};

static kgi_uint pll_command(struct pll_state *state);
static void pll_set(struct kgi_display *, kgi_uint);
static kgi_sint pll_best_state(kgi_uint f, enum kgi_tm_cmd cmd,
	struct pll_state *s, const struct pll_limits *pll);

static struct kgi_clock clock =
{
	"Chips & Technologies",		/* manufacturer		*/
	"655xx",			/* model		*/
	GM_ALL,				/* supported modes 	*/
	0,				/* special capabilities	*/
	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
	NULL,				/* Private data hook	*/
	CLK_TYPE_PROG,			/* clock type		*/
	{{
		400000, 80000000,	/* clock range	FIXME!	*/
	}}
};

/*
<<<< enter the PLL parameters here 
<<<< The output frequency is calculated as
<<<< fout = (fref * factor.mul/factor.div * dividend/divisor) >> post-VCO;
<<<< see the clock chip data sheet for the appopriate ranges and values.

Note: The divider (PSN) can be either 1 or 4. I've enhanced the code to take 
      this into account.
*/

static const struct pll_limits dclk_pll = 
{
	14318180,			/* fref				*/
	{ 4, 1 },			/* factor (divider = 1 || 4)	*/
	{ 48000000, 220000000 },	/* fvco range			*/
	{ 0, 5 },			/* post-VCO divider range	*/
	{ 3, 127 },			/* dividend range		*/
	{ 3, 127 }			/* divisor range		*/
};

#define	INVALID_COMMAND	((kgi_uint) -1)

static kgi_uint pll_command(struct pll_state *s)
{
	return ((s->mul&0x7F)<<11) | ((s->div&0x7F)<<4) | (s->p<<1) | s->psn;
}

static void pll_set(struct kgi_display *dpy, kgi_uint clock_cmd)
{
	/* done in chipset driver */
}

/*
**	KGI module interface
*/

/*	find the best paramter set to produce a certain frequency f
	
	[modified for the 655xx chips]
*/

static kgi_sint pll_best_state(kgi_uint f, enum kgi_tm_cmd cmd,
	struct pll_state *s, const struct pll_limits *pll)
{
	kgi_ratio	ratio;
	struct pll_state states[2];	/* one for each psn */
	struct pll_state *st;
	int		i,j;

	printk("pll_best_state: enter - clock: %i\n",f);

	ASSERT(0 < pll->fvco.min);
	ASSERT(pll->fvco.min <= pll->fvco.max);

	ASSERT(pll->p.min <= pll->p.max);

	ASSERT(0 < pll->mul.min);
	ASSERT(pll->mul.min <= pll->mul.max);

	ASSERT(0 < pll->div.min);
	ASSERT(pll->div.min <= pll->div.max);

	/*	clear state
	*/
	memset(s, 0, sizeof(*s));
	memset(&states, 0, sizeof(states));

	for (i=0;i<=1;i++) {

		j=i ? 4 : 1;		/* psn to try */
		st=&(states[i]);

		/*	determine post-VCO divider/multiplier
		*/
		if ((f < pll->fvco.min) || (pll->fvco.max < f)) {
	
			if (f < pll->fvco.min) {

				states[i].p = (pll->p.min < 0) ? 0 : pll->p.min;
				states[i].fvco = f << st->p;

				while ((st->fvco < pll->fvco.min) &&
					(st->p < pll->p.max)) {

					st->p++;
					st->fvco = f << st->p;
				}
	
			} else {
	
				st->p = (pll->p.max > 0) ? 0 : pll->p.max;
				st->fvco = f >> (- st->p);

				while ((st->fvco > pll->fvco.min) &&
					(st->p > pll->p.min)) {

					st->p--;
					st->fvco = f >> (- st->p);
				}
			}
	
			if ((st->fvco < pll->fvco.min) || (pll->fvco.max < st->fvco) ||
				(st->p < pll->p.min) || (pll->p.max < st->p)) {

				DEBUG1("fvco = %i Hz, p = %i is out of range!", \
					st->fvco, st->p);
				return -EINVAL;
			}

		} else {

			st->fvco = f;
			st->p = 0;
		}

		ratio.mul = st->fvco * pll->a.div * j;  /* psn !! */
		ratio.div = pll->fref * pll->a.mul;	

		ASSERT(ratio.mul > 0);
		ASSERT(ratio.div > 0);

		/*	avoid integer overflow
		*/
		while ((ratio.mul > 0x00100000) && (ratio.div > 0x00100000)) {
	
			ratio.mul >>= 1;
			ratio.div >>= 1;
		}
		if (0 == ratio.mul  ||  0 == ratio.div) {

			DEBUG0("bad parameter set!!");
			return -EINVAL;
		}

		/*	now determine the best mul and div values.
		*/
		st->div = st->mul = -1;
		if ((pll->mul.max - pll->mul.min) < (pll->div.max - pll->div.min)) {

			kgi_sint best_err = 0x7fffffff;
			kgi_sint mul = pll->mul.min;
			kgi_sint div = ((ratio.div * mul) / ratio.mul) +
			(cmd == CMD_LOWER ? 1 : 0);

			while (mul < pll->mul.max) {

				register kgi_sint err = (cmd == CMD_LOWER)
					? div * ratio.mul - mul * ratio.div
					: mul * ratio.div - div * ratio.mul;


				if (0 <= err && (err < best_err)) {

					best_err = err;
					st->mul = mul;
					st->div = div;
				}

				mul++;
				div = ((ratio.div * mul) / ratio.mul) +
					(cmd == CMD_LOWER ? 1 : 0);

				if (pll->div.max < div) {

					break;
				}
			}
		
		} else {

			kgi_sint best_err = 0x7fffffff;
			kgi_sint div = pll->div.min;
			kgi_sint mul = ((ratio.mul * div) / ratio.div) +
				(cmd == CMD_LOWER ? 0 : 1);

			while (div < pll->div.max) {

				register kgi_sint err = (cmd == CMD_LOWER)
					? div * ratio.mul - mul * ratio.div
					: mul * ratio.div - div * ratio.mul;

				if ((0 <= err) && (err < best_err)) {

					best_err = err;
					st->mul = mul;
					st->div = div;
				}

				div++;
				mul = ((ratio.mul * div) / ratio.div) +
					(cmd == CMD_LOWER ? 0 : 1);

				if (pll->mul.max < mul) {

					break;
				}
			}
		}

		if (st->mul < pll->mul.min  ||  pll->mul.max < st->mul ||
			st->div < pll->div.min || pll->div.max < st->div) {
	
			DEBUG1("parameters out of range");
			return -EINVAL;
		}

		st->f = pll->fref * st->mul / st->div;
		st->f = st->f * pll->a.mul / pll->a.div / j;  /* psn !! */
		st->f = (st->p >= 0)
			? st->f >> st->p : st->f << (-st->p);

	}  /* psn loop */
		
	/* Now compare the results of both psns used. */
	if (abs(states[0].f-f) < abs(states[1].f-f)) {
		memcpy(s,&states[0],sizeof(*s));
		s->psn=0;
	} else {
		memcpy(s,&states[1],sizeof(*s));
		s->psn=1;
	}

	printk("fout = %i, fvco = %i, mul = %i, div = %i, p = %i, psn = %i\n", \
		s->f, s->fvco, s->mul, s->div, s->p, s->psn);
		
	printk("pll_best_state: exit\n");	
		
	return EOK;
}

void chips_calcclock(int clock, ChipsRegs *regs) {
	struct pll_state state;

	pll_best_state(clock,CMD_LOWER, &state, &dclk_pll);

	regs->vclock[0]=(state.p << 1) + state.psn;
	regs->vclock[1]=state.mul-2;
	regs->vclock[2]=state.div-2;
}

/*
**	generic KGI module interface
*/

kgi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME DRIVER_REV);

	card->clock = &clock;

	return EOK;
}

void kgim_clock_done(struct kgi_display *dpy)
{
	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}

kgi_sint kgim_clock_command(struct kgi_graphic *state, kgi_uint cmd, kgi_uint arg)
{
	return -E(CLOCK, UNKNOWN);
}

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

kgi_sint kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
	enum kgi_tm_cmd *cmd)
{
	struct pll_state state;
	kgi_uint newfreq = (*cmd == CMD_PROPOSE) 
		? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALID_COMMAND;
		return EOK;

	case CMD_LOWER:

		if (newfreq < MIN_FREQ) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too low",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq > MAX_FREQ) {

			newfreq = MAX_FREQ;
		}

		if (pll_best_state(newfreq, *cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_RAISE:

		if (MAX_FREQ < newfreq) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too high", \
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq < MIN_FREQ) {

			newfreq = MIN_FREQ;
		}

		if (pll_best_state(newfreq, *cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_CHECK:

		mode->clock = INVALID_COMMAND;
		if (pll_best_state(newfreq, CMD_LOWER, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->clock = pll_command(&state);
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}

void kgim_clock_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	if (mode->clock != INVALID_COMMAND) {

		pll_set(dpy, mode->clock);

	} else {

		INTERNAL_ERROR;
	}
}

